/* ***********************************************************
hi_sync  Version 0.1.1
ローカルファイルバックアップ用ライブラリ

	_Dir : ファイルリスト所得クラス (need hi_files)
	C_bakSync : バックアップ実行クラス

2014-04-12
hi_tag	http://studyhall.webcrow.jp/
*********************************************************** */
library hi_sync;
import 'dart:core';
import 'dart:io';
import '../../hi_files/lib/hi_files.dart';
// import 'package:hi_files/hi_files.dart';

const CTRLKEY = 'hi_sync_type';
/**
 * Control file information class
 *
 * dir : control file exist directory
 * type : control type 0 = (reserved), 1 = don't delete
 */
class _ctrlDir {
	String keyword;
	List<String> dir = new List();
	List<int> type = new List();
	_ctrlDir(this.keyword);
}

/**
 * Get file list and properties
 *
 *  Constructor
 *  	_Dir(this.basedir)  <lib:hi_files>
 *
 * 	Properties
 * 		option:	path of under the basedir
 *    flg: 0 = no use, 1 = checked, 2 = copied, 3 = updated, 4 = new, 5 = deleted
 *    type: 0 = directory, 1 = nomal file, 3 = control file, 4 = reserved(ignore), 5 = don't delete
 */
class _Dir extends C_Files {
	String basedir;					// 比較のベースディレクトリ
//	_ctrlDir _ctrl= new _ctrlDir(CTRLKEY);
	_ctrlDir _ctrl= new _ctrlDir('hi_sync_type');

	_Dir(this.basedir) {
		getList(basedir);

		// ファイルを順次走査し、ペースファイル以下のセットと、コントロールファイルの抜き取りを行う
		for (int i = 0; i < filename.length; i++) {
			// control file の処理
			// control file　があったら、そのコントロールタイプとディレクトリを記録する
			if (type[i] == 1) {
				String f = filename[i].substring(0, filename[i].length - 1);
				// _debugprint(i, f);
				if (f == _ctrl.keyword) {
					_ctrl.dir.add(directory[i]);	// 対象directoryを登録
					// control file 名の最後の一文字が control type
					if (filename[i].endsWith('0')) {
						_ctrl.type.add(0);
					} else if (filename[i].endsWith('1')) {
						_ctrl.type.add(1);
					}
					type[i] = 3;
					flg[i] = 1;
					continue;
				}
			}
			// option にベースファイル以下のパスをセットする
			String buff = directory[i];
			buff = buff.substring(basedir.length, buff.length);  // sBaseDir 以下を向き取り
			if (type[i] == 1) {
				buff += ('\\' + filename[i]);		// File の場合
			}
			option[i] = buff;
			flg[i] = 0;
		}

		// _ctrl.dirのチェック
		// type
		//	0(サブディレクトリを含む全てのファイルを無視)
		//	1(サブディレクトリを含む全てのファイルを削除しない)
		for (int j = 0; j < _ctrl.dir.length; j++) {
			for (int i = 0; i < filename.length; i++) {
				if (type[i] == 1 && flg[i] == 0) {
					// 未処理ファイルのみ対象
					if (directory[i].startsWith(_ctrl.dir[j])) {
						switch (_ctrl.type[j]) {
							case 0:
								type[i] = 4;	// ignore
								break;
							case 1:
								type[i] = 5;	// don't delete
								break;
						}
					}
				}
			}
		}
	}
	_debugprint(int i, String s) {
		print(i);
		print(directory[i]);
		print(filename[i]);
		print('type=' + type[i].toString());
		print('debug:' + s);
	}
}

/**
 * Make local files backup
 *
 *  Constructor
 *  	C_bakSync(this.copypath, this.backpath, [this.logfile])
 *  Metohds
 *  	int sync()
 *  		return: 0 = no error, -1 = Exception, -2 = other error
 *  Properties
 *  	String copypath: path of orignal files
 *  	String backpath: path of backup space
 *  	C_log log: log messages and save log file <lib:hi_files>
 */
class C_bakSync {
	String copypath;
	String backpath;
	String logfile;

	C_bakSync(this.copypath, this.backpath, [this.logfile]) {
		if (logfile == null) logfile = 'baksync.log';
	}
	int sync() {
		int iErr = 0;
		String errMessage;
		_Dir copyfiles = new _Dir(copypath);
		_Dir backfiles = new _Dir(backpath);
		_oneDrive oneDrive = new _oneDrive(copypath, backpath);
		C_log log = new C_log(logfile);

		String buff;	// 一時処理用　log message にも使用
		// copyfiles を順次走査
		for(int i = 0; i < copyfiles.flg.length; i++) {
			try {
				// Microsoft OneDrive 対策
				if (oneDrive.chk(copyfiles.directory[i]) || oneDrive.chk(copyfiles.filename[i])) {
					continue;
 				}
				int n = backfiles.option.indexOf(copyfiles.option[i]);
				if (n == -1) {
					// 存在しないとき、コピー
					// control file があるとき、コピーしないようにするには事前に確認が必要(未作成)
					switch (copyfiles.type[i]) {
						case 0:
							// Directory
							buff = backfiles.basedir + copyfiles.option[i];
							log.add('Create:' + buff);		// Error発生時に何をしていたか分かるように先にログを記録
							Directory targetDir = new Directory(buff);
							targetDir.createSync();
							break;
						case 1:
							// File
							File targetfile = new File(copyfiles.basedir + copyfiles.option[i]);
							buff = backfiles.basedir + copyfiles.option[i];
	    				log.add('Copy:' + buff);
	    				targetfile.copySync(buff);
	    				break;
	    			default:
	    				// backfilesに追加
	    				backfiles.option.add(copyfiles.option[i]);
	    				backfiles.flg.add(4);	// 新規作成
	    				copyfiles.flg[i] = 2; // copy 済み
	    					// bacfilesにファイルがないからコピーしているから、すぐに削除はありえないので、type=5は考えない、
					}
				} else {
					// Back先に同一ファイルまたはディレクトリが存在するとき
					backfiles.flg[n] = 1;	// 存在する
					if ((copyfiles.type[i] == 1)
							&& (backfiles.timeStamp[n].isBefore(copyfiles.timeStamp[i]))) {
						// Back先ファイルのほうが古いとき
						File targetfile = new File(copyfiles.basedir + copyfiles.option[i]);
						buff = backfiles.basedir + backfiles.option[n];
						log.add('Update:' + buff);
						targetfile.copySync(buff);
						copyfiles.flg[i] = 2;	// copy済
						backfiles.flg[n] = 3;	// update 済み
					} else {
						copyfiles.flg[i] = 1; // 処理済　(backup先には何もしない)
					}
				}
				// print('copyfile ' + copyfiles.option[i] + '/' +copyfiles.flg[i].toString());
			} on Exception catch(e) {
        errMessage = e.toString();
        log.add(errMessage);
        iErr = -1;
  		} catch(e) {
        errMessage = e.toString();
        log.add(errMessage);
        iErr = -2;
  		}
		}

		// backfiles を走査 (ディレクトリ内のファイルを消去してからディレクトリを消すため逆に走査)
		for(int i = backfiles.flg.length - 1; i >= 0; i--) {
			try {
				// Microsoft OneDrive 対策
				if (oneDrive.chk(backfiles.directory[i]) || oneDrive.chk(backfiles.filename[i])) {
					continue;
 				}
				switch (backfiles.flg[i]) {
					case 0:		// 何も処理していない
						switch (backfiles.type[i]) {
							case 0: // Directory  copy元に存在しないから削除
								buff = backfiles.basedir + backfiles.option[i];
								Directory targetDir = new Directory(buff);
								log.add('Delete:' + buff);
								targetDir.deleteSync(recursive: false);
								backfiles.flg[i] = 5;	// 削除
								break;
							case 1: // File  copy元に存在しないから削除
								buff = backfiles.basedir + backfiles.option[i];
								File targetfile = new File(buff);
		    				log.add('Delete:' + buff);
		    				targetfile.deleteSync(recursive: false);
								backfiles.flg[i] = 5;	// 削除
		    				break;
							case 5: // don't delete copy元にないのでbackからコピーする
								File targetfile = new File(backfiles.basedir + backfiles.option[i]);
								buff = copyfiles.basedir + backfiles.option[i];
		    				log.add('Copy:' + buff);
								targetfile.copySync(buff);
								// copyfiles　に option と flg を追加
		    				copyfiles.option.add(backfiles.option[i]);
		    				copyfiles.flg.add(4);	// 新規作成
  	    				backfiles.flg[i] = 2; // copy 済み
  	    				break;
						}	// flg = 0 の処理end
						break;
					case 1:		// copy元に存在するが何も処理していない
										// backの方が新しければ、backからcopyする
						int n = copyfiles.option.indexOf(backfiles.option[i]);
						switch (backfiles.type[i]) {
							case 1:	// normal file
							case 5:	// don't delete
								if ((n != -1)
    								&& (copyfiles.timeStamp[n].isBefore(backfiles.timeStamp[i]))) {
      						// Back先ファイルのほうが新しいとき
      						File targetfile = new File(backfiles.basedir + backfiles.option[i]);
      						String buff = copyfiles.basedir + copyfiles.option[n];
      						log.add('Update:' + buff);
      						targetfile.copySync(buff);
      						copyfiles.flg[n] = 3;	// update 済
      						backfiles.flg[i] = 2;	// copy 済
    						}
								break;
						}
						break;
				}	// end of switch
			} on Exception catch(e) {
        errMessage = e.toString();
        log.add(errMessage);
        iErr = -11;
  		} catch(e) {
        errMessage = e.toString();
        log.add(errMessage);
        iErr = -12;
  		}
		}	// end of for
		int iLogErr = log.write();
		if (iLogErr != 0) log.add('Log:' + log.errMessage);
		iErr = iLogErr * 100 + iErr;
		return iErr;
	}
}

/**
 * OneDrive 対策  (Ver 0.0.1 から追加)
 * パスに OneDriveが含まれているとき、chk(s)　の　s の一文字目が . のとき true を返す
 */
class _oneDrive {
	String copypath;
	String backpath;
	bool ans;

	_oneDrive(this.copypath, this.backpath) {
		ans = copypath.contains('OneDrive');
		if (!ans) {
			ans = backpath.contains('OneDrive');
		}
	}
	bool chk(String s) {
		bool b;
		if (s == null) {
			b = false;
		} else if (ans && s.length > 2) {
			if (s.substring(0,1) == '.') {
				b = true;
			} else {
				b = false;
			}
		}	else {
			b = false;
		}
		return b;
	}
}